%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















:- ensure_loaded(typeops).

:- module_interface(csppo).

:- import(prolog_types).

% 
% Interface to CSP over partially ordered sets
%

:- type order1(T) --> T =< T.

:- type poset(_T).  % abstract, a partially ordered set

:- type special_type --> bottom;top.

%----
% po_create( +Domain, +Order, -PO) :-
%       PO is the representation of partial order over Domain.
%
:- pred po_create(list(@T),list(order1(T)),poset(@T)).


%----
% po_extend( +PO, +Ids, -POnew) :-
%       Ids are added to domain of PO,
%       result is POnew.
%
 :- pred po_extend(poset(T),list(T),poset(@T)).


%----
% po_test_leq( +PO, +X, +Y) :- 
%	(X =< Y) is true in PO.
%
:- pred po_test_leq(poset(@T),T,T).


%----
% po_lub(PO,X,Y,Z) :- 
% 	Z is least upper bound of X and Y in partial order PO,
%	fails if no upper bound exists or lub is not unique.
%
:- pred po_lub(poset(@T),T,T,T).


%----
% po_maximal( + PO, + X) :-
%	Element X is maximal in the order of PO
%
:- pred po_maximal(poset(@T),T).


%----
% po_minimal( + PO, + X) :-
%       Element X is minimal in the order of PO
%
:- pred po_minimal(poset(@T),T).


%----
% po_glb( +PO, +X, +Y, ?Z) :-
%       Z is greatest lower bound of X and Y in partial order PO,
%       fails if no lower bound exists or glb is not unique.
%
:- pred po_glb(poset(@T),T,T,T).


%----
% po_write(PO) :- 
%       writes out the domain and order relations of PO
%
:- pred po_write(poset(_)).


%---- 
% po_indomain(PO,X) :- 
%       Element X is element in the domain of PO 
% 
:- pred po_indomain(poset(@T),T).


%---- 
% po_ordered(PO,X) :- 
%       Element X is element in the order of PO 
% 
:- pred po_ordered(poset(@T),T).


%----
% po_mub(PO,X,Y,Z) :- Z is minimal upper bound of X and Y.
%
:- pred po_mub(poset(@T),T,T,T).


%----
% po_less(PO,X,Y) :- (X =< Y) and X \= Y is true in PO.
%
:- pred po_less(poset(@T),T,T).


%----
% three kinds of constraints are handled
%
:- type po_constraint(T) -->
        T = T;
        T =< T;
        lub(T,list(T)).   % X = S \lub Y \lub ...
                          % X = S + Y + ... in external syntax

%----
% satisfy_constraints_once( +PO, +C) :-
%       constraints C are satisfied in poset PO,
%       PO must be a quasi-lattice.
%       C may contain logical variables when called.
%	On success, variables in C are bound to satisfying values.
%       only the first solution is computed.
%
:- pred satisfy_constraints_once(poset(@T),list(po_constraint(T))).

%----
% satisfy_constraints( +PO, +C) :-
%       constraints C are satisfied in poset PO,
%       PO must be a quasi-lattice.
%       C may contain logical variables when called.
%	On success, variables in C are bound to satisfying values.
%
:- pred satisfy_constraints(poset(@T),list(po_constraint(T))).

:- end_interface(csppo).











