%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















:- ensure_loaded(typeops).

:- import(prolog_types).
:- import(prolog_meta).
:- import(prolog_io).

:- import(lib).
:- import(glib).
:- import(files).
:- import(decl).
:- import(csppo).
:- import(titrace).

:- type typename.

:- type typename > modtypename.

:- type modtypename > simpletypename.

:- type simpletypename --> {X|simple(X)}.        % A kind of hack!

:- type modtypename --> filename:typename.

:- type predname.

:- type predname > modpredname.

:- type modpredname > aritypredname.

:- type aritypredname > simplepredname.

:- type simplepredname --> {X|simple(X)}.        % A kind of hack!

:- type aritypredname --> predname/posint.

:- type modpredname --> filename:aritypredname.

%:- type constructor --> (top;top);{top|goal};{top^goal}, top.      % prelimary definition

:- type constructor --> top.      % prelimary definition

:- type constrdef --> c(typename,constructor).

:- type equivdef --> typename == typename.

:- type subdef --> typename < typename.

%:- type predkind --> pred;predl;pred_neg. % defined in decl.pl

:- type constraints == atom.   % prelimary definition

:- type preddef --> p(modpredname,list(typename),predkind);p(modpredname,list(typename),predkind,constraints).

:- type clause == goal.

:- type exportstatement --> ep(aritypredname);et(typename);eat(typename).

:- type importstatement --> i(filename);im(filename);ip(filename,aritypredname);it(filename,typename);iat(filename,typename).

:- pred pragma_option(top).
:- dynamic pragma_option/1.

:- pred io_opened_module(filename).
:- dynamic io_opened_module/1.

:- pred io_read_module(filename).
:- dynamic io_read_module/1.

:- pred module_decsription(moduledescription).
:- dynamic module_description/1.

:- type typesub --> typename =< typename.
:- pred type_sub(typesub).
:- dynamic type_sub/1.

:- pred type_equiv_map(typename, typename).
:- dynamic type_equiv_map/2.

:- pred tap(top).

tap(['-v'|Args]) :-
        assert(pragma_option(verbose)),!,
        tap(Args).

tap(['-dummy-pdecl'|Args]) :-
        assert(pragma_option(dummy_pdecl)),!,
        tap(Args).
 
tap(['-dummy-fdecl'|Args]) :-
        assert(pragma_option(dummy_fdecl)),!,
        tap(Args).
 
tap([]) :-
        write(end),nl,
        halt.
 
tap([FileName|FileList]) :- 
        atom(FileName), !, 
        write('tap: checking '),write(FileName),nl,
	read_module(FileName, _Exports, Types, Constructors, Equivs, Subs, Preds, Clauses),
	filter_modulename(Types,CTypes),
        % add built-in type goal, must be know for valid_type_names	
%write(CTypes),nl,write(Subs),nl,write(Constructors),nl,write(Preds),nl,write(Clauses),nl,
	process_equivs(Equivs),
	check_alias_decls(Subs, Constructors),
	process_subtypes(Subs, Ordering),
	process_constructors(Constructors),
	process_predicates([goal|CTypes],Preds),
	process_input([goal|CTypes],Ordering,Clauses),
        retract_everything,
        tap(FileList).
 
tap([File|_]) :-
        write('tap>>  invalid file name '),
        write(File), 
        write(' .halt'), nl, halt.
 
tap(File) :- 
        tap([File]).
        
:- pred check_type_decls(list(subdef), list(constrdef)).

check_alias_decls(Os, Co) :-
	check_alias_order(Os),
	check_alias_constructor(Co).

:- pred check_alias_order(list(subdef)).

check_alias_order([]).
check_alias_order([X < Y | OL]) :-
	(type_equiv_map(X,_Type1) -> 
            (write('type alias '),write(X), 
	     write(' used in subtype relationship '), write(X < Y),
	     nl, halt);
	    (type_equiv_map(Y,_Type2) -> 
		(write('type alias '), write(Y), 
		 write(' used in subtype relationship '), write(X < Y),
		 nl, halt);
	        check_alias_order(OL))).

:- pred check_alias_consructor(list(constrdef)).

check_alias_constructor([]).
check_alias_constructor([c(Type, Constructor) | CL]) :-
	(type_equiv_map(Type,_Type1) -> 
            (write('type alias '), write(Type),
	     write(' also defined through '), write(Type),
	     write(' --> '), write(Constructor),
	     nl, halt);
	     check_alias_constructor(CL)).

% predicates are treated specially with sp_member
:- pred special_remove_duplicates(list(preddef),list(preddef)).

special_remove_duplicates([],[]).
special_remove_duplicates([X|Xs],Ys) :-
        sp_member(X,Xs), !,
        special_remove_duplicates(Xs,Ys).
special_remove_duplicates([X|Xs],[X|Ys]) :-
        % not sp_member(X,Xs),
        special_remove_duplicates(Xs,Ys).

:- pred sp_member(preddef,list(preddef)).
 
sp_member(X,[Y|_]) :- X = Y,!.
% double predicate declarations may have different variables for parameters 
% and constraints nevertheless they are equal
sp_member(p(Module:PName/Arity, _Args1, PredKind),[p(Module:PName/Arity, _Args2, PredKind)|_]).
sp_member(p(Module:PName/Arity, _Args1, PredKind, _Co1),[p(Module:PName/Arity, _Args2, PredKind, _Co2)|_]).

sp_member(X,[_|T]) :- sp_member(X,T).


:- pred retract_everything.
 
retract_everything :-
        retractall(pred_decl_variant_ngr0(_,_,_,_,_)),
        retractall(func_decl_variant_ngr0(_,_,_,_)), 
        retractall(par_num_counter(_)), 
        retractall(pragma_option(_)), 
        retractall(type_equiv_map(_,_)), 
        retractall(type_sub(_)), 
	retractall(module_description(_)),
        retractall(io_opened_module(_)), 
        retractall(io_read_module(_)), 
        retractall(tinf_diag(_)).

:- pred filter_modulename(list(modtypename),list(simpletypename)).

filter_modulename([],[]).
filter_modulename([_Module:Type|TL1],[Type|TL2]) :-
	filter_modulename(TL1,TL2).

/*filter_abstract([],[]).
filter_abstract([abstract(Mod:Type)|TypeList],[Mod:Type|NewTypeList]) :-
	!,
	retractall(func_decl_variant_ngr0(_Funcname,_Arity,_ArgTypes,Type)),
	retractall(type_sub(_Type1 =< Type)),
	!,
	filter_abstract(TypeList,NewTypeList).
filter_abstract([Type|TypeList],[Type|NewTypeList]) :- 
	!,
filter_abstract(TypeList,NewTypeList).
*/
:- pred process_equivs(list(equivdef)).

process_equivs([]).
process_equivs([Type == RealType|Rest]) :-
	assert(type_equiv_map(Type,RealType)),!,
	process_equivs(Rest).
	
:- pred process_subtypes(list(subdef)).

process_subtypes([],[]).
process_subtypes([Type1 < Type2|Rest], [FX =< FY|NewRest]) :-
	normalize_type_equiv(Type1,Xe),
	normalize_type_equiv(Type2,Ye),
	Xe =..[FX|ArgsX],
        Ye =..[FY|ArgsY],
        ArgsX == ArgsY,    % parameters must be in same sequence 
        !,
	process_subtypes(Rest,NewRest).
process_subtypes([Type1 < Type2|_Rest],_) :-
        write('tap: warning: subtyping '), write(Type1<Type2), write(' not allowed.'),
        fail.

:- pred process_constructors(list(constrdef)).

process_constructors([]).
process_constructors([c(Type,Constructor)|Rest]) :-
	func_transform(Type,Constructor),
	process_constructors(Rest).

:- pred process_predicates(list(typename),list(preddef)).

process_predicates(_Types,[]).
process_predicates(Types,[p(_Module:PName/Arity, Args, PredKind)|Rest]) :-
       \+ test_pred_decl_variant_ngr(PName,Arity,_,_),  % must be unique
        !,
        trans_pdecl_list(Args,Args1,Co1),
        normalize_type_equiv_list(Args1,Args2),
        valid_type_names_list(Types,Args2),
        normalize_type_equiv_list(Co1,Co2),
        valid_type_names_co(Types,Co2),
        simplify_pdecl(Args2,Co2,Co3),
        de_store_pdecl(PName,Arity,PredKind,Args2,Co3),!,
	process_predicates(Types, Rest).
process_predicates(_Types,[p(_Module:PName/Arity, _Args1, _PredKind)|_Rest]) :-
        write('tap: error:  multiple declaration for predicate '), 
	write(PName/Arity),nl,halt.
process_predicates(Types,[p(_Module:PName/Arity, Args1, PredKind,Co1)|Rest]) :-
	\+ test_pred_decl_variant_ngr(PName,Arity,_,_),  % must be unique
        !,
        commas_to_l(Co1,Co2),
        trans_pdecl_list(Args1,Args2,Co3),
        append(Co2,Co3,Co4),
        normalize_type_equiv_list(Args2,Args3),
        valid_type_names_list(Types,Args3),
        normalize_type_equiv_list(Co4,Co5),
        valid_type_names_co(Types,Co5),
        simplify_pdecl(Args3,Co5,Co6),
        de_store_pdecl(PName,Arity,PredKind,Args3,Co6),!,
	process_predicates(Types, Rest).

process_predicates(_Types,[p(_Module:PName/Arity, _Args1, _PredKind,_Co)|_Rest]) :-
        write('tap: error:  multiple declaration for predicate '), 
	write(PName/Arity),nl,halt.

%
% valid_type_names(Symbols,TypeTerm) :-
%       every functor in TypeTerm is member of Symbols.
%
:- pred valid_type_names(list(typename),top).

valid_type_names(_Ss,Term) :-
        var(Term), !.
valid_type_names(_Ss,top) :- !.
valid_type_names(Ss,Term) :-
        % \+ var(Term),
        Term =.. [F|Args],
        memberchk_open(F,Ss), 
        !,
        valid_type_names_list(Ss,Args).
valid_type_names(Ss,Term) :-
        nl,
        write('type symbol unknown '), write(Term), nl,
        write('not in ':Ss),nl,
        halt.
 
:- pred valid_type_names_list(list(typename),list(top)).

valid_type_names_list(_Ss,[]).
valid_type_names_list(Ss,[X|Xs]) :-
        valid_type_names(Ss,X),
        valid_type_names_list(Ss,Xs).
 
:- pred valid_type_names_co(list(typename),list(top)).

valid_type_names_co(_Ss,[]). 
valid_type_names_co(Ss,[X=<Y|C]) :-
        valid_type_names(Ss,X),
        valid_type_names(Ss,Y),
        valid_type_names_co(Ss,C).
 
% Process the function symbols on the Rhs of  type Lhs --> Rhs
% Insert their types in DB.

:- pred func_transform(typename,constructor).
 
func_transform(Type, (Cons1;Cons2)) :- !,
        f_transform(Type, Cons1), 
        func_transform(Type,Cons2).
 
func_transform(Type, Cons1) :- 
	f_transform(Type,Cons1).
 
% Process a function symbol 
% 
:- pred f_transform(typename,constructor).

f_transform(Type,Term) :- 
        Term = {Var|Goal},
        de_store_fdecl_goal(Type,Var,Goal).
 
f_transform(Type,Term) :- 
        Term = {Var^Goal},
        de_store_fdecl_goal(Type,Var,Goal).
 
f_transform(Type,Term) :- 
        Term =.. [Functor | ArgList],
        % type_preserving(Functor,ArgList,Type),  not required !?
        !,
        functor(Term,_,Arity),
        normalize_type_equiv_list(ArgList,Args3),
        de_store_fdecl(Functor,Arity,Args3,Type).
 
f_transform(_Type,Term) :- 
        Term =.. [Functor | _ArgList],
        write('tap: error: function symbol '), write(Functor), 
        write(' is not type preserving; Cannot continue '), nl, halt.
 
:- pred process_input(list(typename),list(typesub),list(clause)).

process_input(Ts,Os,Cls) :- !,
	close_list(Ts), close_list(Os), close_list(Cls),
        % add 'xx', often chosen as arbitrary instantiation in var-Typing
        po_create([xx|Ts],Os,PO),
        true,
        true,
        ((func_decl_variant_ngr0(X1,Xn,X2,X3),
            true,fail) ; true),
        ((pred_decl_variant_ngr0(X1,Xn,_Xp,X2,X3),
            true,fail) ; true),
        true,
        type_infer_cl(PO,Cls).
/*
        Lat = PO,
        (option_with_bottom -> complete_bottom(Lat,TO1); TO1=Lat),
        (pragma_option(with_top) -> 
            complete_top(TO1,TO2)
        ;   TO2 = TO1
        ),
        type_infer_cl(TO2,Cls).
*/
 
% 

:- pred read_module(filename,list(exportstatement),list(modtypename),list(constrdef),list(equivdef),list(subdef),list(preddef),list(clause)).

read_module(FileName, Exports, Types, Constructors, Equivs, Subs, Preds, Clauses) :-
	file_dealiased(tap(FileName),File),
        absolute_file_name(File,AbsName),
        io_read_module(AbsName),!,      % already read
	module_description(module(AbsName, Exports, Types, Constructors, Equivs, Subs, Preds, Clauses)).

read_module(FileName, _Exports, _Types, _Constructors, _Equivs, _Subs, _Preds, _Clauses) :-
	file_dealiased(tap(FileName),File),
        absolute_file_name(File,AbsName),
        io_opened_module(AbsName),!,      % already open
	write('tap: error: cycle in import-statements! Can not continue!'), nl,
	halt.

read_module(FileName, Exports, Types2, Constructors2, Equivs2, Subs2, Preds2, Clauses) :-
	file_dealiased(tap(FileName),File),
        absolute_file_name(File,AbsName),
	assert(io_opened_module(AbsName)),
	(pragma_option(verbose) ->
	 (write(read_module(FileName)), nl) ;
	    true),
	open(AbsName,read,Stream),
	% divide file in different lists
	read_module2(Stream,module(AbsName,DT,TC,TE,TS,DP,Im,Ex,Clauses)),
	close(Stream),
	close_list(DT),close_list(TC),close_list(TE),
	close_list(TS),close_list(DP),close_list(Im),
	close_list(Ex),close_list(Clauses),
	% mark types and predicates with modulename
	mark_type_list(AbsName,DT,MDT),
	mark_pred_list(AbsName,DP,MDP),
	% import everything from importlist Im
	import_module_list(Im,IDT,ITC,ITE,ITS,IDP,IEX),
	append(MDT, IDT, Types),
	append(MDP, IDP, Preds),
	append(TC, ITC, Constructors),
	append(TE, ITE, Equivs),
	append(TS, ITS, Subs),!,
	% remove duplicate Types, Equivs, ... after import
	remove_duplicates(Types, Types2),
	remove_duplicates(Equivs, Equivs2),
	remove_duplicates(Subs, Subs2),
	remove_duplicates(Constructors, Constructors1),
	special_remove_duplicates(Preds, Preds2),
	% trueif Types and Constructors aren't multiple defined
	check_types(Types2),
	check_constructors(Constructors1, Constructors2),
	% have all exports been defined?
	append(Ex, IEX, Exports),
        check_exports(Exports, Types2, Preds2),
        %write(module(AbsName, Exports, Types2, Constructors2, Equivs2, Subs2, Preds2, Clauses)),nl,nl,
	% save all data in DB 
	assert(module_description(module(AbsName, Exports, Types2, Constructors2, Equivs2, Subs2, Preds2, Clauses))),
	retract(io_opened_module(AbsName)),
	assert(io_read_module(AbsName)).

:- pred check_types(list(typename)).

check_types([]).
check_types([Filename:Type|Rest]) :-
	(member(_Other_Filename:Type,Rest)  ->
	 (write('tap: error: multiple declaration of type '),write(Type),
	  write(' in file '),write(Filename),write(' and file '),
	  write(_Other_Filename),write(' !'),nl,halt);
	check_types(Rest)).

:- pred check_constructors(list(constrdef), list(constrdef)).

check_constructors([],[]).
check_constructors([c(Type,_Constructor) | Rest],NewRest) :-
	member(c(Type, _X), Rest),!,
	check_constructors(Rest,NewRest).

check_constructors([CDef|Rest], [CDef|NewRest]) :-
	check_constructors(Rest, NewRest).
	
:- pred mark_type_list(filename, list(typename), list(modtypename)).

mark_type_list(_,[],[]).
mark_type_list(FileName,[Element|List],[FileName:Element|NewList]) :-
	mark_type_list(FileName,List,NewList).

:- pred mark_pred_list(filename, list(preddef), list(preddef)).
	
mark_pred_list(_,[],[]).
mark_pred_list(FileName,[p(PName/Arity, Args, Predkind)|List],[ p(FileName:PName/Arity, Args, Predkind)|NewList]) :-
	mark_pred_list(FileName,List,NewList).
mark_pred_list(FileName,[p(PName/Arity, Args, Predkind, Co)|List],[ p(FileName:PName/Arity, Args, Predkind, Co)|NewList]) :-
	mark_pred_list(FileName,List,NewList).
	
% arbeite die Importliste ab.
:- pred import_module_list(list(importstatement),list(modtypename),list(constrdef),list(equivdef),list(subdef),list(preddef),list(exportstatement)).

import_module_list([],[],[],[],[],[],[]).
import_module_list([Importstatement|Imports], Types, Constructors, Equivs, Subs, Preds, Exports) :-
	import_module(Importstatement,DT,TC,TE,TS,DP,EX),
	import_module_list(Imports,IDT,ITC,ITE,ITS,IDP,IEX),
	append(DT,IDT,Types),
	append(TC,ITC,Constructors),
	append(TE,ITE,Equivs),
	append(TS,ITS,Subs),
	append(DP,IDP,Preds),
	append(EX,IEX,Exports).

% lade das File, werte die Exportschnittstelle und den Typ Importanweisung aus 
% und gib die gefilterten Ergebnisse zurueck.
  
:- pred import_module(importstatement,list(modtypename),list(constrdef),list(equivdef),list(subdef),list(preddef),list(exportstatement)).


% simple import of all exported types and predicates
import_module(i(FileName), Types, Constructors, Equivs, Subs, Preds, []) :-
	read_module(FileName, EX, DT, TC, TE, TS, DP, _CL),
	filter_exports(EX, EX, DT, TC, TE, TS, DP, Types, Constructors, Equivs, Subs, Preds).

% simple import of all exported types and predicates plus generating exportstatements
import_module(im(FileName), Types, Constructors, Equivs, Subs, Preds, EX) :-
	read_module(FileName, EX, DT, TC, TE, TS, DP, _CL),
	filter_exports(EX, EX, DT, TC, TE, TS, DP, Types, Constructors, Equivs, Subs, Preds).
	
% import of a single predicate
import_module(ip(FileName, Id), [], [], [], [], [Pred], []) :-
	read_module(FileName, EX, _DT, _TC, _TE, _TS, DP, _CL),
	(member(ep(Id),EX) -> 
	 (member(p(Mod:Id, Args,Predkind),DP) -> 
	  Pred = p(Mod:Id, Args, Predkind);
	  (member(pc(Mod:Id, Args, Predkind, Co),DP) ->
	   Pred = pc(Mod:Id, Args, Predkind, Co);
	   (write('tap: error: predicate '), write(Id),
            write(' is exported but not defined by module '), write(FileName),nl,halt)));
	 (write('tap: error: predicate '), write(Id),
	  write(' is not exported by module '), write(FileName), nl, halt)).
% hier koennten aus Args noch die Typen herausgeholt werden
% dann muesste die Parameterliste geaendert werden!
% import of a single type 	
import_module(it(FileName, Id), Types, Constructors, Equivs, Subs, [], []) :-
	read_module(FileName, EX, DT, TC, TE, TS, _DP, _CL),
	(get_type(Id, EX) ->
	 export_type(et(Id),EX,DT,TC,TE,TS,Types,Constructors, Equivs, Subs);
	 (write('tap: error: type '), write(Id),
	  write(' is not exported by module '), write(FileName), nl, halt)). 
	
% import a type abstractly only fails if type isn't exported by other module  
import_module(iat(FileName, Id), [Mod:Id], [], [], [], [], []) :-
	read_module(FileName, EX, DT, _TC, _TE, _TS, _DP, _CL),
	(get_type(Id,EX) ->
	 member(Mod:Id,DT);
	 (write('tap: error: abstract type '), write(Id),
	  write(' is not exported by module '), write(FileName), nl, halt)). 

:- pred get_type(typename, list(exportstatement)).

% is Id exported abstract or normal by a module? 	
get_type(top,_).   % top is global    
get_type(goal,_).  % goal is global, too
get_type(Id,EX) :-
	member(et(Id),EX).
get_type(Id,EX) :-
	member(eat(Id),EX).
	
:- pred filter_exports(list(exportstatement),list(exportstatement),list(modtypename),list(constrdef),list(equivdef),list(subdef),list(preddef),list(modtypename),list(constrdef),list(equivdef),list(subdef),list(preddef)).

filter_exports([], _, _, _, _, _, _, [], [], [], [], []).

filter_exports([ep(ID)|EX], EXB, DT, TC, TE, TS, DP, Types, Constructors, Equivs, Subs, Preds):-
	export_pred(ID, EXB, DT, TC, TE, TS, DP, DT1, TC1, TE1, TS1, DP1), 
	filter_exports(EX, EXB, DT, TC, TE, TS, DP, DT2, TC2, TE2, TS2, DP2),
	append(DT1, DT2, Types),
	append(TC1, TC2, Constructors),
	append(TE1, TE2, Equivs),
	append(TS1, TS2, Subs),
	append(DP1, DP2, Preds).

filter_exports([ExpStatement|EX], EXB, DT, TC, TE, TS, DP, Types, Constructors, Equivs, Subs, Preds):-
	export_type(ExpStatement, EXB, DT, TC, TE, TS, DT1, TC1, TE1, TS1), 
	filter_exports(EX, EXB, DT, TC, TE, TS, DP, DT2, TC2, TE2, TS2, Preds),
	append(DT1, DT2, Types),
	append(TC1, TC2, Constructors),
	append(TE1, TE2, Equivs),
	append(TS1, TS2, Subs).

:- pred export_pred(aritypredname,list(exportstatement),list(modtypename),list(constrdef),list(equivdef),list(subdef),list(preddef),list(modtypename),list(constrdef),list(equivdef),list(subdef),list(preddef)).

% code to extract types in predicate definition should be included here

export_pred(ID, _EXB, _DT, _TC, _TE, _TS, DP, [], [], [], [], [p(Mod:ID, Args, Predkind)]) :-
	member(p(Mod:ID, Args, Predkind), DP), ! .

export_pred(ID, _EXB, _DT, _TC, _TE, _TS, DP, [], [], [], [], [p(Mod:ID, Args, Predkind, Co)]) :-
	member(p(Mod:ID, Args, Predkind, Co), DP), ! .

:- pred export_type(exportstatement,list(exportstatement),list(modtypename),list(constrdef),list(equivdef),list(subdef),list(modtypename),list(constrdef),list(equivdef),list(subdef)).

export_type(et(ID), EX, DT, _TC, TE, _TS, [Mod:ID], [], [ID == RealType], []) :-
	member(ID == RealType, TE), 	
	member(Mod:ID, DT),
	compound(RealType),!,
	RealType =.. [FrontType|_Args],
	(get_type(FrontType, EX) ->
	 true;
	 (write('tap: error: type equivalence '), write(ID == FrontType),
	  write('in module '), write(Mod),
	  write(' is not exportable because '), write(FrontType),
	  write(' is not exported by '), write(Mod), nl, halt)).

export_type(et(ID), EX, DT, _TC, TE, _TS, [Mod:ID], [], [ID == RealType], []) :-
	member(ID == RealType, TE), !, 	
	member(Mod:ID, DT),
	(get_type(RealType, EX) ->
	 true;
	 (write('tap: error: type equivalence '), write(ID == RealType),
	  write('in module '), write(Mod),
	  write(' is not exportable because '), write(RealType),
	  write(' is not exported by '), write(Mod), nl, halt)).

/*
export_type(et(ID), EX, DT, TC, TE, TS, [Mod:ID|DT1], TC3, TE1, TS1) :-
	member(Mod:ID, DT),
	(setof(X, member(X < ID, TS), Exportlist) ->
	 (export_type_list(Exportlist, EX, DT, TC, TE, TS, DT1, TC1, TE1, HTS1),
	  setof(X < ID, member(X < ID, TS), HTS2),
	  append(HTS1, HTS2, TS1));
	 DT1 = [], TC1 = [], TE1 = [], TS1 = []),
	(setof(c(ID, CO), member(c(ID, CO), TC), TC2) ->
	 append(TC1, TC2, TC3);
	 TC3 = TC1).	    
*/
export_type(et(ID), EX, DT, TC, TE, TS, [Mod:ID|DT1], TC3, TE1, TS1) :-
	member(Mod:ID, DT),
	(setof(X, member(X < ID, TS), Exportlist) ->
	 (export_type_list(Exportlist, EX, DT, TC, TE, TS, DT1, TC1, TE1, HTS1),
	  setof(X < ID, member(X < ID, TS), HTS2),
	  append(HTS1, HTS2, TS1));
	 DT1 = [], TC1 = [], TE1 = [], TS1 = []),
	(setof(c(ID, CO), member(c(ID, CO), TC), TC2) ->
	 append(TC1, TC2, TC3);
	 (compound_member(ID,TC,TC4),
	  append(TC1,TC4,TC3))).

export_type(eat(ID), _EX, DT, _TC, _TE, _TS, [Mod:ID], [], [], []) :-
	member(Mod:ID,DT).	

:- pred compound_member(typename,list(constrdef),list(constrdef)).

compound_member(_,[],[]).

compound_member(ID,[c(Compound,Constraint)|Rest],[c(Compound,Constraint)|NewRest]) :-
	compound(Compound),
	Compound =.. [ID|_],!,
	compound_member(ID,Rest,NewRest).

compound_member(ID,[_|Rest],NewRest) :-
	compound_member(ID,Rest,NewRest).

% export_type_list
%
% export all types from the list of typenames

:- pred export_type_list(list(typename),list(exportstatement),list(modtypename),list(constrdef),list(equivdef),list(subdef),list(modtypename),list(constrdef),list(equivdef),list(subdef)).

export_type_list([], _, _, _, _, _, [], [], [], []).
export_type_list([ID|RestID], EX, DT, TC, TE, TS, Types, Constructors, Equivs, Subs) :-
	 export_type(et(ID), EX, DT, TC, TE, TS, DT1, TC1, TE1, TS1),
	 export_type_list(RestID, EX, DT, TC, TE, TS, DT2, TC2, TE2, TS2),!,
	 append(DT1, DT2, Types),
	 append(TC1, TC2, Constructors),
	 append(TE1, TE2, Equivs),
	 append(TS1, TS2, Subs).

:- pred check_exports(list(exportstatement),list(modtypename),list(preddef)).	
% test if all exported types and predicates are declared

check_exports([],_,_).
check_exports([ep(PName/Arity)|Ex],DT,DP) :-
	member(p(_Mod:PName/Arity, _Args, _Predkind),DP), ! ,
        check_exports(Ex,DT,DP).
check_exports([ep(PName/Arity)|Ex],DT,DP) :-
	member(p(_Mod:PName/Arity, _Args, _Predkind, _Co),DP), ! ,
        check_exports(Ex,DT,DP).
check_exports([et(ID)|Ex],DT,DP) :-
	member(_Mod:ID,DT), ! ,
        check_exports(Ex,DT,DP).
check_exports([eat(ID)|Ex],DT,DP) :-
	member(_Mod:ID,DT), ! ,
        check_exports(Ex,DT,DP).
check_exports([ExStatement|Ex],DT,DP) :-
	ExStatement =.. [_|Arg],
        write('tap: warning: '), 
	write(Arg), write(' is exported, but not declared!'),nl,
	check_exports(Ex,DT,DP).

:- type moduledescription --> module(filename,list(typename),list(constrdef),list(equivdef),list(subdef),list(preddef),list(importstatement),list(exportstatement),list(goal)).

:- pred read_module2(stream, moduledescription).

read_module2(Stream,ModuleDesc) :-
        read(Stream,Cl),
        !,
        expand_term(Cl,Clause),   % handles things such as grammar rules
        !,
        ( Clause == end_of_file -> 
	 (retractall(pragma_option(tap_Module(_))),
	  retractall(pragma_option(tap_Body(_))),
	  retractall(pragma_option(tap_Interface(_))),
	  true);
	 (transform(Clause,ModuleDesc), !,
	  read_module2(Stream,ModuleDesc))).
    
:- pred transform(goal, moduledescription).

transform((:- op(N,A,Op)), _) :- ! , 
	% define operator, otherwise read can't handle new Op's
	op(N,A,Op).

transform((type pragma(Option)), _ModuleDesc) :- !,
	assert(pragma_option(Option)).


% module begin/end interface/body statements

transform((:- module(FileName)), _ModuleDesc) :- !,
	assert(pragma_option(tap_Module(FileName))).

transform((:- module_interface(FileName)), ModuleDesc) :- !,
	assert(pragma_option(tap_Interface(FileName))),
	transform((:- module(FileName)), ModuleDesc).

transform((:- end_interface(FileName)), _ModuleDesc) :- 
	retract(pragma_option(tap_Interface(FileName))),!,
	true. % Exportiere alle nicht lokalen/abstrakten Typen + Preds 

transform((:- end_interface(FileName)), _) :- 
	write('tap: error: no :-'),
	write(module_interface(FileName)),
	write(' found for :- '),
	write(end_interface(FileName)),nl,halt. 

transform((:- module_body(FileName)), ModuleDesc) :- !,
	(pragma_option(tap_Interface(FileName)) ->
	 (write('tap: warning: no :- '),
	  write(end_interface(FileName)),
	  write(' found for :- '),
	  write(module_interface(FileName)),nl, 
	  transform((:- end_interface(FileName))));
	 true),	 
	assert(pragma_option(tap_Body(FileName))),
	transform((:- module(FileName)), ModuleDesc).

transform((:- end_module), ModuleDesc) :- 
	pragma_option(tap_Module(FileName)),!,
	transform((:- end_module(FileName)), ModuleDesc). 

transform((:- end_module), _) :- !,
	write('tap: warning: no :- module(some_name) found for :- end_module. '),nl,halt. 
	
transform((:- end_module(FileName)), _ModuleDesc) :- 
	retract(pragma_option(tap_Body(FileName))),!,
	retract(pragma_option(tap_Module(FileName))),!,
	true. % Exportiere alle erlaubten Typen + Preds 

transform((:- end_module(FileName)), _ModuleDesc) :- 
	retract(pragma_option(tap_Module(FileName))),!,
	true. % Exportiere alle erlaubten Typen + Preds 

transform((:- end_module(FileName)), _) :- !, 
	write('tap: error: no :- '),
	write(module(FileName)),
	write(' found for :- '),
	write(end_module(FileName)),nl,halt. 


% syntactical variants

% This is for compapility with older versions, don't use this anymore!
transform((:- type_import(FileName)), ModuleDesc) :- !,
        transform((:- import_type(FileName)), ModuleDesc).

transform((:- use_module(FileName)), ModuleDesc) :- !,
        transform((:- import(FileName)), ModuleDesc).

transform((:- use_module(FileName, Id)), ModuleDesc) :- !,
        transform((:- import(FileName, Id)), ModuleDesc).

transform((:- module(FileName, Id)), ModuleDesc) :- !,
        transform((:- module(FileName)), ModuleDesc),
        transform((:- export(Id)), ModuleDesc).

transform((:- public(Id)), ModuleDesc) :- !,
	transform((:- export(Id)), ModuleDesc).

transform((:- transfer(FileName, Id)), ModuleDesc) :- !,
        transform((:- import(FileName, Id)), ModuleDesc),
        transform((:- export(Id)), ModuleDesc).

% import statements

transform((:- import(_,[])),_) :- !.
transform((:- import(FileName, [Id|More])),module(_Name,_DT,_TC,_TE,_TS,_DP,Im,Ex,_Cl)) :- !,
        (pragma_option(tap_Interface(_FileName)) ->
	  insert_end(e(Id), Ex);
	  (pragma_option(tap_Module(_FileName)) ->
	   true;
	   insert_end(e(Id), Ex))),
	insert_end(ip(FileName, Id), Im),
	transform((:- import(FileName, More)),module(_Name,_DT,_TC,_TE,_TS,_DP,Im,Ex,_Cl)).

transform((:- import(FileName)), module(_Name,_DT,_TC,_TE,_TS,_DP,Im,_Ex,_Cl)) :- !,
        (pragma_option(tap_Interface(_FileName)) ->
	  insert_end(im(FileName), Im);
	  (pragma_option(tap_Module(_FileName)) ->
	   insert_end(i(FileName), Im);
	   insert_end(im(FileName), Im))).
	
transform((:- import(FileName, Id)), module(_Name,_DT,_TC,_TE,_TS,_DP,Im,Ex,_Cl)) :- !,
        (pragma_option(tap_Interface(_FileName)) ->
	  insert_end(e(Id), Ex);
	  (pragma_option(tap_Module(_FileName)) ->
	   true;
	   insert_end(e(Id), Ex))),
	insert_end(ip(FileName, Id), Im).
	
transform((:- import_type(FileName)), module(_Name,_DT,_TC,_TE,_TS,_DP,Im,_Ex,_Cl)) :- !,
	% momentarily the same as :- import(Filename).
        (pragma_option(tap_Interface(_FileName)) ->
	  insert_end(im(FileName), Im);
	  (pragma_option(tap_Module(_FileName)) ->
	   insert_end(i(FileName), Im);
	   insert_end(im(FileName), Im))).
		
transform((:- import_type(FileName, abstract(Id))), module(_Name,_DT,_TC,_TE,_TS,_DP,Im,Ex,_Cl)) :- !,
	(pragma_option(tap_Interface(_FileName)) ->
	 insert_end(eat(Id), Ex);
	 (pragma_option(tap_Module(_FileName)) ->
	  true;
	  insert_end(eat(Id), Ex))),
	insert_end(iat(FileName, Id), Im).

transform((:- import_type(_,[])),_) :- !.
transform((:- import_type(FileName, [Id|More])),module(_Name,_DT,_TC,_TE,_TS,_DP,Im,Ex,_Cl)) :- !,
	(pragma_option(tap_Interface(_FileName)) ->
	 insert_end(et(Id), Ex);
	 (pragma_option(tap_Module(_FileName)) ->
	  true;
	  insert_end(et(Id), Ex))),
	insert_end(it(FileName, Id), Im),
	transform((:- import_type(FileName, More)),module(_Name,_DT,_TC,_TE,_TS,_DP,Im,Ex,_Cl)).
transform((:- import_type(FileName, Id)), module(_Name,_DT,_TC,_TE,_TS,_DP,Im,Ex,_Cl)) :- !,
	(pragma_option(tap_Interface(_FileName)) ->
	 insert_end(et(Id), Ex);
	 (pragma_option(tap_Module(_FileName)) ->
	  true;
	  insert_end(et(Id), Ex))),
	insert_end(it(FileName, Id), Im).

	
% export statements
transform((:- export([])),_) :- !.
transform((:- export([Id|More])),module(_Name,_DT,_TC,_TE,_TS,_DP,_Im,Ex,_Cl)) :- !,
        insert_end(ep(Id), Ex),
	transform((:- export(More)),module(_Name,_DT,_TC,_TE,_TS,_DP,_Im,Ex,_Cl)).
 
transform((:- export(Id)), module(_Name,_DT,_TC,_TE,_TS,_DP,_Im,Ex,_Cl)) :- !,
        insert_end(ep(Id), Ex).
	
transform((:- export_type([])),_) :- !.
transform((:- export_type([Id|More])),module(_Name,_DT,_TC,_TE,_TS,_DP,_Im,Ex,_Cl)) :- !,
        insert_end(et(Id), Ex),
	transform((:- export_type(More)),module(_Name,_DT,_TC,_TE,_TS,_DP,_Im,Ex,_Cl)).

transform((:- export_type(abstract(Id))), module(_Name,_DT,_TC,_TE,_TS,_DP,_Im,Ex,_Cl)) :- !,
        insert_end(eat(Id), Ex).
	
transform((:- export_type(Id)), module(_Name,_DT,_TC,_TE,_TS,_DP,_Im,Ex,_Cl)) :- !,
        insert_end(et(Id), Ex).

% change form command to :- command

transform((type Definition), ModuleDesc) :- !,
	transform((:- type Definition), ModuleDesc).
transform((subtype Definition), ModuleDesc) :- !,
	transform((:- subtype Definition),ModuleDesc).
transform((suptype Definition), ModuleDesc) :- !,
	transform((:- suptype Definition), ModuleDesc).
transform((pred Declaration), ModuleDesc) :- !,
	transform((:- pred Declaration), ModuleDesc).
transform((predl Declaration), ModuleDesc) :- !,
	transform((:- predl Declaration), ModuleDesc).
transform((pred_neg Declaration), ModuleDesc) :- !,
	transform((:- pred_neg Declaration), ModuleDesc).
transform((rule Declaration), ModuleDesc) :- !,
	transform((:- rule Declaration), ModuleDesc).

% more syntactical sugar

transform((:- type Type = RealType),  ModuleDesc) :- !,
	transform((:- type Type == RealType),  ModuleDesc).

transform((:- type Type1 < Type2), ModuleDesc) :- !,
	transform((:- subtype Type1 < Type2), ModuleDesc).
transform((:- suptype Type1 > Type2), ModuleDesc) :- !,
	transform((:- subtype Type2 < Type1), ModuleDesc).
transform((:- type Type1 > Type2), ModuleDesc) :- !,
	transform((:- subtype Type2 < Type1), ModuleDesc).

% subtype relationship

transform((:- subtype Type1 < Type2), module(_Name,_DT,_TC,_TE,TS,_DP,_Im,_Ex,_Cl)) :- !,
        insert_end(Type1 < Type2, TS).

% type equivalences

transform((:- type Type == RealType), module(_Name, DT,_TC,TE,_TS,_DP,_Im,Ex,_Cl)) :- !,
	store_type(Type, DT,Ex),
	insert_end(Type == RealType, TE).

%  type constructor declarations and function type declarations

transform((:- type Type --> Constructors), module(_Name, DT,TC,_TE,_TS,_DP,_Im,Ex,_Cl)) :- !,
	store_type(Type, DT,Ex),
        insert_end(c(Type, Constructors), TC).

%  just type constructors : no function symbols
transform((:- type Type), module(_Name, DT,_TC,_TE,_TS,_DP,_Im,Ex,_Cl)) :- !,
	store_type(Type, DT,Ex).

:- pred store_type(typename, list(typename), list(exportstatement)).

store_type(Type, DT, Ex) :-
	check_type_name(Type),
	functor(Type,TypeName,Arity),
	Arity > 0,!,
	insert_end(TypeName, DT),
	(pragma_option(tap_Interface(_FileName)) ->
	  insert_end(et(TypeName), Ex);
	 (pragma_option(tap_Module(_FileName)) ->
	  true;
	  insert_end(et(TypeName), Ex))).
	

store_type(Type, DT, Ex) :-
        check_type_name(Type),
	insert_end(Type, DT),
	(pragma_option(tap_Interface(_FileName)) ->
	 insert_end(et(Type), Ex);
	 (pragma_option(tap_Module(_FileName)) ->
	  true;
	  insert_end(et(Type), Ex))).	     

% predicate type declaration

transform((:- pred {Pred}), ModuleDesc) :- !,
	store_pdecl(ModuleDesc, Pred,pred).
transform((:- predl {Pred}), ModuleDesc) :- !,
	store_pdecl(ModuleDesc, Pred,predl).
transform((:- pred Pred -> Co), ModuleDesc) :- !,  % style variant
	transform((pred Pred with Co), ModuleDesc).
transform((:- predl Pred -> Co), ModuleDesc) :- !,  % style variant
	transform((predl Pred with Co), ModuleDesc).
transform((:- pred Pred with Co), ModuleDesc) :- !,
	store_pdecl(ModuleDesc, Pred,pred,Co).
transform((:- predl Pred with Co), ModuleDesc) :- !,
	store_pdecl(ModuleDesc, Pred,predl,Co).
transform((:- pred_neg Pred | Co), ModuleDesc) :- !,
	store_pdecl(ModuleDesc, Pred,pred_neg,Co).
transform((:- pred_neg Pred with Co), ModuleDesc) :- !,
	store_pdecl(ModuleDesc, Pred,pred_neg,Co).
transform((:- pred Pred), ModuleDesc) :- !,
	store_pdecl(ModuleDesc, Pred,pred).
transform((:- predl Pred), ModuleDesc) :- !,
	store_pdecl(ModuleDesc, Pred,predl).
transform((:- pred_neg Pred), ModuleDesc) :- !,
	store_pdecl(ModuleDesc, Pred,pred_neg).

% grammar rule

transform((:- rule NonTerminal), ModuleDesc) :- !, 
	NonTerminal =.. [F|Args],
	append(Args,[list(top),list(top)],A2),
	Pred =.. [F|A2],
	store_pdecl(ModuleDesc, Pred, predl). 

:- pred store_pdecl(moduledescription, goal, predkind).

store_pdecl(module(_Name, _DT,_TC,_TE,_TS,DP,_Im,Ex,_Cl), Pred, Predkind) :-
	Pred =.. [PName|Args],
	functor(Pred,PName,Arity),
	insert_end( p(PName/Arity, Args, Predkind), DP),
	(pragma_option(tap_Interface(_FileName)) ->
	 insert_end( ep(PName/Arity), Ex);
	 (pragma_option(tap_Module(_FileName)) ->
	  true;
	  insert_end( ep(PName/Arity), Ex))).

:- pred store_pdecl(moduledescription, goal, predkind, constraints).

store_pdecl(module(_Name, _DT,_TC,_TE,_TS,DP,_Im,Ex,_Cl), Pred, Predkind, Co) :-
	Pred =.. [PName|Args],
	functor(Pred,PName,Arity),
	insert_end( p(PName/Arity, Args, Predkind, Co), DP),
	(pragma_option(tap_Interface(_FileName)) ->
	 insert_end( ep(PName/Arity), Ex);
	 (pragma_option(tap_Module(_FileName)) ->
	  true;
	  insert_end( ep(PName/Arity), Ex))).
 
% end of file: 

transform(end_of_file, _ModuleDesc) :- !,
	write('eof should be handled elsewhere').

% else it is a program clause

transform(Clause, module(_Name, _DT,_TC,_TE,_TS,_DP,_Im,_Ex,Cl)) :- 
	insert_end(Clause, Cl), !.

% or something strange!

transform(Clause, _) :- 
	write('tap: error: can not transform '), 
	write(Clause), nl,
	halt.

% insert in open list.

:- pred insert_end(T, list(T)).

insert_end(X, L) :- var(L), !, L = [X|_NewL].
insert_end(X, [_|L]) :- insert_end(X, L).

% close open list.

:- pred close_list(list(T)).

close_list([]) :- !.
close_list([_|L]) :- close_list(L).


:- pred check_type_name(typename).

check_type_name(bottom) :- 
        % type bottom is forbidden
        write('type bottom is forbidden, use another typename.'),
        !,fail.    
check_type_name(Type) :-
        % parameters must be pairwise distinct variables.
        Type =.. [_F|Args],
        pairwise_distinct_var(Args),
        !.
check_type_name(Type) :-
        write(Type:'parameters must be pairwise distinct variables.'),
        fail.

:- pred pairwise_distinct_var(list(T)).

pairwise_distinct_var([]).
pairwise_distinct_var([X|Xs]) :-
        var(X),
        \+ var_member(X,Xs),
        pairwise_distinct_var(Xs).

:- pred normalize_type_equiv(T,list(T)).     

normalize_type_equiv(Term,T9) :- 
        var(Term), !,
        T9 = Term.
normalize_type_equiv(Term,T9) :-
        % \+ var(Term),
        type_equiv_map(Term,T2), !,
        normalize_type_equiv(T2,T9).
normalize_type_equiv(Term,T9) :-
        % \+ var(Term),
        % \+ type_equiv_map(Term,T2), !,
        Term =.. [F|Args],
        normalize_type_equiv_list(Args,Args2),
        T9 =.. [F|Args2].
 
:- pred normalize_type_equiv_list(list(typename),list(goal)).     

normalize_type_equiv_list([],[]).
normalize_type_equiv_list([X|Xs],[Y|Ys]) :-
        normalize_type_equiv(X,Y),
        normalize_type_equiv_list(Xs,Ys).

:- pred commas_to_l(top,list(T)).

commas_to_l((A,B),[A|BL]) :- 
	!, commas_to_l(B,BL).
commas_to_l(A,[A]).
