%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















:- ensure_loaded(typeops). 

:- module_interface(ti_cl).
% 
%  The main predicate ti_cl/3 takes a type order and a set of marked
%  atoms which have to be type-checked. For each atom the corresponding
%  type declaration is looked up and appropriate type constraints
%  (principal upper matcher constraints and other ordering constraints)
%  are generated.
%

:- import(prolog_types).
:- import(prolog_num).
:- import(prolog_io).
:- import(prolog_meta).

:- import(checking).
:- import(glib).
:- import(csppo).
:- import(ttlib).
:- import(cllib).
:- import(ttord).
:- import(ti_2po).
:- import(ti_tt).  % type defs for constraints
:- import(titrace).

:- op(600,xfy,(#)).  % s.a. ti_clause.pl

% given Reduct,Reduct,...
% generate CSP :  VarTyping, Matcher, Constraints

%  no bottom
	
:- type special_ptyp --> 
	o ;      % any logical atom,  o can be read as 'boolean type'
	g.       % marks a generic head atom

:- type builtin_type  --> goal.

:- type latom = top.          % a logical atom 

:- type reduct -->
	 latom # special_ptyp ;
	 term : typeterm ;         % term has type
	 term * typeterm ;         % term has principal type
	 stop.  % for internal use

:- type head_constr = po_constraint(top).

:- type clc --> clc( poset(top),    % type order
		  list(reduct),     % Reduct
		  list(tt_pum),   % PUMs
		  list(head_constr),     % head - constraints
		  list(tt_co1),     % constraints over type terms
		  list(top)).    % bottom-vars, 


%----
% ti_cl( +PO, +Red, -CLC) :-
%	PO is the given type order,
%	Red represents the set of atoms that will be type-checked.
%	Red is a list of terms of type 'reduct' where all terms have the form
%	'latom # special_ptyp'.
%	Red is reduced, i.e. decomposed, into 
%	principal upper matchers and other constraints,
%	The result is stored in the structure LC.
%
% ti_cl may be non-deterministic if function symbols are overloaded
%
:- pred ti_cl(poset(_),list(reduct),clc).
%
ti_cl(PO,Red,CLC) :-
        append(Red,[stop],RedS),
	ti_clb(PO, RedS,[],[],[],[], CLC).



% old
:- pred ti_cl(poset(_),list(reduct),list(tt_pum),list(head_constr),list(tt_co1),clc).
ti_cl(Lat, Red,PUMs,Ch,Co, CLC) :- 
        append(Red,[stop],RedS),
	ti_clb(Lat, RedS,PUMs,Ch,Co,[], CLC).

% normalize_var_type(Lat,Red,VT,Co) :- normalize Red into VT + Co.

% X:Type is not transformed if X occurs only once
%
:- pred normalize_var_type(poset(_),list(reduct),list(reduct),list(tt_co1)).  
normalize_var_type(_Lat,[],[],[]) :- !.
normalize_var_type(Lat,[ValVar:VarType|Red],VT,Co) :-
% v:t1, v:t2 -> v:glb(t1,t2)
        true, 
        member_elim(ValVar:VT2,Red,Red2), !,
        true,
	true,
        true,
        glbtt(Lat, VarType,VT2,VTglb),
        true,
        normalize_var_type(Lat, [ValVar:VTglb|Red2], VT,Co).
normalize_var_type(Lat,[ValVar:VarType|Red], VT,[VT2 =< VarType|Co]) :-
% v:t1, v * t2 ->  v*t2, t2 =< t1
        true, true,
        member(ValVar*VT2,Red), !,
        true,
        normalize_var_type(Lat, Red, VT,Co).
normalize_var_type(Lat,[ValVar*VarType|Red], VT,[VarType =< VT2|Co]) :-
% v*t1, v : t2 ->  v*t1, t1 =< t2
        true,
        member_elim(ValVar:VT2,Red,Red2), !,
        true,
        true,
        normalize_var_type(Lat, [ValVar*VarType|Red2], VT,Co).
normalize_var_type(Lat,[ValVar*VarType|Red], VT,Co) :-
% v*t1, v * t2 ->  v*t2, t1 = t2
        true,
        member(ValVar*VT2,Red), !,
        true,
        VarType = VT2,
        normalize_var_type(Lat, Red, VT,Co).
normalize_var_type(Lat,[Vt|Red], [Vt|Red2],Co) :-  % var of Vt only once now
        normalize_var_type(Lat, Red,Red2,Co).

:- end_interface(ti_cl).

:- module_body(ti_cl).

:- pred ti_clb(poset(_),list(reduct),list(tt_pum),list(head_constr),list(tt_co1),list(_),clc).
%
ti_clb(_Lat, Red,PUMs,_Ch,Co,B, _CLC) :- 
	ti_trace_flag, nl,
	write('type-infer'),nl,
	%write('    lat:'),po_write(Lat),nl,
	write('    red':Red),nl,
	write('    pums':PUMs),nl,
	%write('    ch':Ch),nl,
	write('    co':Co),nl,
	write('    bv':B),nl,
	get0(_Xx),
	fail.


ti_clb(Lat, [Head#g|Red],PUMs,_Ch0,Co,B, CLC) :-
        Head =.. [Pred|Args], !,
        true,
        true,
	functor(Head,_,Arity),
        pred_decl_generic(Pred,Arity,Domain,Ch,Ids),  
	true,
        true,
        true,  
 	true, true,
        make_f_pairs(:,Args,Domain,ArgsRed),
        append(ArgsRed,Red,Red2),
        % add_all_typeIds_from_head_to_poset  !!!
        po_extend(Lat,Ids,Lat2),
        ti_clb(Lat2, Red2,PUMs,Ch,Co,B, CLC).


ti_clb(Lat, [GoalT|Red],PUMs,Ch,Co,B, CLC) :-
	GoalT = Goal#o ,
	value_var(Goal),
	!,
	ti_clb(Lat, [Goal*goal|Red],PUMs,Ch,Co,B, CLC).


ti_clb(Lat, [Atom:GoalType|Red],PUMs,Ch,Co,B, CLC) :-
        \+ value_var(Atom),
        \+ var(GoalType),
        GoalType = goal,
        !,
        true,
	ti_clb(Lat, [Atom#o|Red],PUMs,Ch,Co,B, CLC).

ti_clb(Lat, [_Term:TermType|Red],PUMs,Ch,Co,B, CLC) :-
        ground(TermType),
        TermType = top,
        !,
        true,
	% take it for granted
        ti_clb(Lat, Red,PUMs,Ch,Co,B, CLC).

 
ti_clb(Lat, [Term:TermType|Red],PUMs,Ch,Co,B, CLC) :-
        \+ value_var(Term),
        Term =.. [Func|Args], !,
        true,
        true,
        % with bottom:  func_decl_variant_ngr_bot(D,Func,Domain,Target),
	functor(Term,_,Arity),
	func_decl_variant_ngr_botvar(Func,Arity,Domain,Target,BotVars),
        % overloaded functions symbols by backtracking
	(true
        ;true,fail),
        true,
        true,
        greatest_lower_match(Lat, Target,TermType),
        true,
	% if function declaration is transparent, then Domain is ground
	(ground(Domain) -> true
         ;  write('Error: declaration not transparent':Func)),
        make_type_pairs(Args,Domain,ArgsRed),
        append(ArgsRed,Red,Red2),
	append(B,BotVars,B2),
        ti_clb(Lat, Red2,PUMs,Ch,Co,B2, CLC).
 

% ti_clb(Lat, [Atom#o|Red],PUMs,Ch,Co,B, CLC) :-
ti_clb(Lat, [Goal#o|Red],PUMs,Ch,Co,B, CLC) :-
	\+ value_var(Goal),
	!,
	true,
	Goal =.. [Pred|Args],
	functor(Goal,_,Arity),
	pred_decl_variant_ngr(Pred,Arity,Domain,Cp),   % non-ground repres.
	%gen_logvar_list(Args,Tvars),
	%make_ltype_pairs(Args,Tvars,ArgsRed),
	%make_lequm_multipairs(Tvars,Domain,ArgsM),
	% ti_cl_pdecl is replaces previous gen_log*, and make_*pairs
	ti_cl_pdecl(Args,Domain,ArgsRed,ArgsM),
	append(ArgsRed,Red,Red2),
	append(PUMs,ArgsM,PUMs2), % nicht [ArgsM]
	append(Co,Cp,Co2),
	ti_clb(Lat, Red2,PUMs2,Ch,Co2,B, CLC).

% generate term:type, term*type, and matchers where appropriate
%ti_cl_pdecl(Args,Domain,ArgsRed,ArgsM).
% realizes optimization for ground domain-types
%
:- pred ti_cl_pdecl(list(term),list(typeterm),list(reduct),list(tt_pum)).
ti_cl_pdecl([],[],[],[]).
ti_cl_pdecl([A|As],[D|Ds],[A:D|ArgsRed],ArgsM) :-
	ground(D),
	!,
	ti_cl_pdecl(As,Ds,ArgsRed,ArgsM).
ti_cl_pdecl([A|As],[D|Ds],[A*Dnew|ArgsRed],[pum(pum_p,[Dnew=<=D])|ArgsM]) :-
%                           type error     [pum(pum_p,[Dnew=<=D]),ArgsM]
	true,
	ti_cl_pdecl(As,Ds,ArgsRed,ArgsM).


ti_clb(Lat, [GoalT|Red],PUMs,Ch,Co,B, CLC) :-
        % keep in mind: GoalT may be Term*TypeVar
        GoalT = Goal*goal,   % built-in type 'goal'
        \+ value_var(Goal),
        % no cut, GoalT may have been Term*TypeVar
        true,
        Goal =.. [Pred|Args],
	functor(Goal,_,Arity),
        test_pred_decl_variant_ngr(Pred,Arity,Domain,Cp),   % non-ground repres.
        %  may backtrack into next clause later
        % now same as above:
        %gen_logvar_list(Args,Tvars),
        %make_ltype_pairs(Args,Tvars,ArgsRed),
        %make_lequm_multipairs(Tvars,Domain,ArgsM),
	ti_cl_pdecl(Args,Domain,ArgsRed,ArgsM),
        append(ArgsRed,Red,Red2),
        append(PUMs,ArgsM,PUMs2), % fehler, nicht [ArgsM]
        append(Co,Cp,Co2),
        ti_clb(Lat, Red2,PUMs2,Ch,Co2,B, CLC).
 

ti_clb(Lat, [Term*TermType|Red],PUMs,Ch,Co,B, CLC) :-
	\+ value_var(Term),
	Term =.. [Func|Args], 
        !,
	true,
	functor(Term,_,Arity),
	func_decl_variant_ngr_botvar(Func,Arity,Domain,Target,BotVars),
	true,
	true,
	TermType = Target,
	%gen_logvar_list(Args,Tvars),
	%make_ltype_pairs(Args,Tvars,ArgsRed),
	%make_lequm_pairs(Tvars,Domain,ArgsM),
	ti_cl_fdecl(Args,Domain,ArgsRed,ArgsM),
	true,
	true,
	append(ArgsRed,Red,Red2),
	true,
	(   ArgsM == [] ->
	    PUMs2 = PUMs
	;  
	    append(PUMs,[pum(pum_f,ArgsM)],PUMs2) 
	),
	%     append(PUMs,[pum(pum_f,[ArgsM]),PUMs2), 
	%                 type-error
	% type-error! append(PUMs,ArgsM,PUMs2),
	append(B,BotVars,B2),
	ti_clb(Lat, Red2,PUMs2,Ch,Co,B2, CLC).

 
% generate term:type, term*type, and matchers where appropriate
%ti_cl_fdecl(Args,Domain,ArgsRed,ArgsM).
% realizes optimization for ground domain-types 
%
:- pred ti_cl_fdecl(list(term),list(typeterm),list(reduct),list(pum_constr(typeterm))).
ti_cl_fdecl([],[],[],[]).
ti_cl_fdecl([A|As],[D|Ds],[A:D|ArgsRed],ArgsM) :-
        ground(D),
        !,
        ti_cl_fdecl(As,Ds,ArgsRed,ArgsM).
ti_cl_fdecl([A|As],[D|Ds],[A*Dnew|ArgsRed],[Dnew=<=D|ArgsM]) :-
%                           type error     [Dnew=<=D,ArgsM]
        true,
        ti_cl_fdecl(As,Ds,ArgsRed,ArgsM).
 


% rotate first element
ti_clb(Lat, [R|Red],PUMs,Ch,Co,B, CLC) :-
	\+ R=stop, !,  
	true,
        true,
	append(Red,[R],Red2),
	ti_clb(Lat, Red2,PUMs,Ch,Co,B, CLC).

ti_clb(Lat, [stop|Red],PUMs,Ch,Co,B, CLC) :- true,
	CLC = clc(Lat,Red,PUMs,Ch,Co,B).


:- end_module(ti_cl). 
