%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















:- ensure_loaded(typeops). 

:- module_interface(tinf_more).

%
%  type inference working on  clc-structure
%

:- ensure_loaded(typeops). 

:- ensure_loaded(ti_2po). 

:- import(prolog_types).
:- import(prolog_num).
:- import(prolog_meta).
:- import(glib).
:- import(csppo).
:- import(ti_2po).
:- import(ti_clause).  % task
:- import(ti_cl).

:- pred option_with_bottom.
option_with_bottom :- fail.
% s.a. tinf.pl tap.pl

% type_inf_more( +CLC, -C, -VT) :-
%	CLC is a structure containing a type-inference task, which already
%	has been preprocessed by ti_clause and ti_cl.
%	Result is a set of constraints C, such that VT is a valid
%	variable typing and CLC (hence the input clause) is type-consistent.
%       fails if CLC is not type-consistent
:- pred type_inf_more(clc,list(top),list(top)).
type_inf_more(CLC,C,VT) :-	
	CLC = clc(Lat2,Red,PUMs,HC,Co,BotVars),
	true,  % 
	true,
	normalize_var_type(Lat2,Red,VT0,Cadd),
	true,
	append(Co,Cadd,Co2),
	true,
	true,
	true,
	explode_simpum(PUMs,PUMse),
	true,
	explode_C(Co2,Co2e),
	true,
	decompose_simpum_co(Lat2,BotVars, PUMse,Co2e,_BotVars2, PCo1,Co3),
	% no cut !,  function symbols may be overloaded
        % ? constrain to minimal_types(Lat2,_BotVars2), 
	true,
	true,true,
	true,
	true,
        true,
        true,
	simplify_simpum_constraints(PCo1,PCo),
        true,
	simplify_constraints(Lat2,Co3,CCo),
	true,
	true,
        true,
        true,
	simpum2lub(PCo,PLub),
        true,
        true,
	simplify_constraints(Lat2,PLub,SPLub),
        true,
        true,
	true,
	satisfy_plub_co(Lat2,SPLub,CCo),
	true,
	valid_vt(Lat2,VT0),
	%    true,
        % ?? minimal_types(Lat2,_BotVars2), 
	true,
	C = HC, implode_C(VT0,VT),
	true.
% now: true

%---
% simpum2lub(PUMs,C) :- 
%       simultaneous p.u.m. PUMs is transformed into set of constraints C 
% 
:- pred simpum2lub(list(pum(T)),list(po_constraint(T))). 
% 
simpum2lub([],[]). 
simpum2lub([pum(_,L)|PUMs],Cs) :-
        make_lub_constraints(L,C), 
        append(C,Cmore,Cs),    % should be real diff.list  
        simpum2lub(PUMs,Cmore). 
  
:- pred simplify_simpum_constraints(list(pum(T)),list(pum(T))).
%
simplify_simpum_constraints([],[]).
simplify_simpum_constraints([pum(Id,L)|PUMs],[pum(Id,Ls)|S]) :-
	simplify_pum_constraints(L,Ls),
	simplify_simpum_constraints(PUMs,S).
 
:- pred simplify_pum_constraints(list(pum_constr(@T)),list(pum_constr(T))).
% 
simplify_pum_constraints([],[]).
simplify_pum_constraints([_X=<=Y|C],C2) :-
        ground(Y), Y=top, !, type_pragma(non_generic),        
        true,
        simplify_pum_constraints(C,C2).
simplify_pum_constraints([Q|C],[Q|C2]) :-
        simplify_pum_constraints(C,C2).

%----
% satisfy_plub_co(PO,PCo,Co) :-
%	try to satisfy constraints PCo from simultaneous p.u.m
%	and constraint Co in type order PO.
% 	PC0 and Co must be in reduced form, i.e., fully decomposed.
%
:- pred satisfy_plub_co(poset(@T),
           list(po_constraint(T)),list(po_constraint(T))).
%
satisfy_plub_co(PO,PLub,Co) :-
	append(PLub,Co,C),
	satisfy_constraints_once(PO,C).


:- end_interface(tinf_more).

:- module_body(tinf_more).

% valid_vt(Lat, VT) :- VT is list of v:typeterm and each typterm not empty.
:- pred valid_vt(poset(_T),list(top)).
valid_vt(_, []).
valid_vt(Lat,[Vt|L]) :- 
	(Vt = _V:T;  Vt = _V*T),
	( \+ground(T) 
        -> % write('warning: valid non-ground type? ':T),nl,
	   numbervars(T,26,_)  % gives A1, B1, C1, ...
           % subst_top(T) ?
        ; true
        ),
	univrecb(T,T0),
	nonempty(Lat,T0),
	valid_vt(Lat,L).

% Instantiations of BotVars must/should be minimal ?
:- pred minimal_types(poset(@T),list(T)).
minimal_types(_PO,[]).
minimal_types(PO,[B|Bs]) :-
	minimal_type(PO,B),
	minimal_types(PO,Bs).

:- pred minimal_type(poset(_T),top).
minimal_type(_PO,X) :- 
	var(X).
minimal_type(PO,X) :-
	X = [_F|Args],
	minimal_types(PO,Args).
minimal_type(PO,X) :-
	true,
	po_minimal(PO,X).

:- end_module(tinf_more).
